# from jaraco.util.classutil
def itersubclasses(cls, _seen=None):
	"""
	itersubclasses(cls)

	Generator over all subclasses of a given class, in depth-first order.

	>>> bool in list(itersubclasses(int))
	True
	>>> class A(object): pass
	>>> class B(A): pass
	>>> class C(A): pass
	>>> class D(B,C): pass
	>>> class E(D): pass
	>>>
	>>> for cls in itersubclasses(A):
	...		print(cls.__name__)
	B
	D
	E
	C
	>>> # get ALL (new-style) classes currently defined
	>>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
	[...'tuple', ...]
	"""

	if not isinstance(cls, type):
		raise TypeError('itersubclasses must be called with '
			'new-style classes, not %.100r' % cls)
	if _seen is None: _seen = set()
	try:
		subs = cls.__subclasses__()
	except TypeError:  # fails only when cls is type
		subs = cls.__subclasses__(cls)
	for sub in subs:
		if sub in _seen:
			continue
		_seen.add(sub)
		yield sub
		for sub in itersubclasses(sub, _seen):
			yield sub



from unittest import mock
import pytest

from hgtools import managers

def test_existing_only():
	"""
	Test the static method RepoManager.existing_only.
	"""
	# presumably, '/' is never an hg repo - at least for our purposes, that's
	#  a reasonable assumption.
	mgrs = managers.RepoManager.get_valid_managers('/')
	existing = list(managers.RepoManager.existing_only(mgrs))
	assert not existing

@mock.patch.object(managers.RepoManager, 'get_valid_managers',
	classmethod(lambda cls, location: iter(())))
def test_no_valid_managers():
	"""
	When no valid managers can be found, a StopIteration is raised providing
	a nice message.
	"""
	with pytest.raises(StopIteration) as err:
		managers.RepoManager.get_first_valid_manager()
	assert 'no source repo' in str(err).lower()



import collections
from distutils.version import StrictVersion

from hgtools import versioning

class VersionedObject(versioning.VersionManagement):
	def __init__(self, **kwargs):
		self.__dict__.update(kwargs)

	def is_modified(self):
		return False

class TestVersioning(object):
	def test_tag_versions(self):
		"""
		Versioning should only choose relevant tags (versions)
		"""
		mgr = VersionedObject(get_tags = lambda: tags)
		tags = set(['foo', 'bar', '3.0'])
		assert mgr.get_tagged_version() == StrictVersion('3.0')
		tags = set([])
		assert mgr.get_tagged_version() == None
		tags = set(['foo', 'bar'])
		assert mgr.get_tagged_version() == None

	def test_tag_priority(self):
		"""
		Since Mercurial provides tags in arbitrary order, the versioning
		support should infer the precedence (choose latest).
		"""
		mgr = VersionedObject(get_tags = lambda: tags)
		tags = set(['1.0', '1.1'])
		assert mgr.get_tagged_version() == '1.1'
		tags = set(['0.10', '0.9'])
		assert mgr.get_tagged_version() == '0.10'

	def test_defer_to_parent_tag(self):
		"""
		Use the parent tag if on the tip
		"""
		mgr = VersionedObject(
			get_tags = lambda rev=None: set(['tip']),
			get_parent_tags = lambda rev=None: set(['1.0']),
		)
		assert mgr.get_tagged_version() == '1.0'

	def test_get_next_version(self):
		mgr = VersionedObject(
			get_repo_tags = lambda: set([])
		)
		assert mgr.get_next_version() == '0.0.1'

	def test_local_revision_not_tagged(self):
		"""
		When no tags are available, use the greatest tag and add the increment
		"""
		mgr = VersionedObject(
			get_tags = lambda rev=None: set([]),
			get_repo_tags = lambda: set(
				collections.namedtuple('tag', 'tag')(var)
				for var in ['foo', 'bar', '1.0'])
		)
		assert mgr.get_tagged_version() is None
		assert mgr.get_next_version() == StrictVersion('1.0.1')
		assert mgr.get_current_version() == '1.0.1.dev0'



import os

import pytest

from hgtools import managers
from hgtools.managers import cmd
from hgtools.managers import subprocess

def test_subprocess_manager_invalid_when_exe_missing():
	"""
	If the hg executable dosen't exist, the manager should report
	False for .is_valid().
	"""
	non_existent_exe = '/non_existent_executable'
	assert not os.path.exists(non_existent_exe)
	mgr = subprocess.GitManager()
	mgr.exe = non_existent_exe
	assert not mgr.is_valid()

@pytest.mark.usefixtures("git_repo", scope='function')
class TestTags(object):
	def setup_method(self, method):
		self.mgr = managers.GitManager('.')

	def teardown_method(self, method):
		del self.mgr

	def test_single_tag(self):
		assert self.mgr.get_tags() == set([])
		self.mgr._invoke('tag', '-am', "Tagging 1.0", '1.0')
		assert self.mgr.get_tags() == set(['1.0'])
		self.mgr._invoke('checkout', '1.0')
		assert self.mgr.get_tags() == set(['1.0'])

class TestParseVersion:
	def test_simple(self):
		assert cmd.Git._parse_version('git version 1.9.3') == '1.9.3'

	def test_trailing_mess(self):
		val = cmd.Git._parse_version('git version 1.9.3 (Mac OS X)')
		assert val == '1.9.3'



import os
import re
import operator
import itertools
import collections

import pkg_resources


TaggedRevision = collections.namedtuple('TaggedRevision', 'tag revision')


class Command(object):
	def is_valid(self):
		try:
			# Check if both command and repo are valid
			self._invoke('status')
		except Exception:
			return False
		return super(Command, self).is_valid()

	def version(self):
		"""
		Return the underlying version
		"""
		lines = iter(self._invoke('version').splitlines())
		version = next(lines).strip()
		return self._parse_version(version)

	@classmethod
	def _parse_version(cls, version):
		return re.search(cls.version_pattern, version).group(1)


class Mercurial(Command):
	exe = 'hg'
	version_pattern = r'Mercurial Distributed SCM \((.*?)\)'

	def find_root(self):
		try:
			return self._invoke('root').strip()
		except Exception:
			pass

	def find_files(self):
		"""
		Find versioned files in self.location
		"""
		all_files = self._invoke('locate', '-I', '.').splitlines()
		# now we have a list of all files in self.location relative to
		#  self.find_root()
		# Remove the parent dirs from them.
		from_root = os.path.relpath(self.location, self.find_root())
		loc_rel_paths = [os.path.relpath(path, from_root)
			for path in all_files]
		return loc_rel_paths

	def get_parent_revs(self, rev=None):
		cmd = ['parents', '--style', 'default',
			'--config', 'defaults.parents=']
		if rev:
			cmd.extend(['--rev', str(rev)])
		out = self._invoke(*cmd)
		cs_pat = '^changeset:\s+(?P<local>\d+):(?P<hash>[0-9a-zA-Z]+)'
		return (match.groupdict()['local'] for match in
			re.finditer(cs_pat, out))

	def get_tags(self, rev=None):
		"""
		Get the tags for the given revision specifier (or the
		current revision if not specified).
		"""
		rev_num = self._get_rev_num(rev)
		# rev_num might end with '+', indicating local modifications.
		return (
			set(self._read_tags_for_rev(rev_num))
			if not rev_num.endswith('+')
			else set([])
		)

	def _read_tags_for_rev(self, rev_num):
		"""
		Return the tags for revision sorted by when the tags were
		created (latest first)
		"""
		return (tr.tag for tr in self._read_tags_for_revset(rev_num))

	def _read_tags_for_revset(self, spec):
		"""
		Return TaggedRevision for each tag/rev combination in the revset spec
		"""
		cmd = ['log', '--style', 'default',  '--config', 'defaults.log=',
			'-r', spec]
		res = self._invoke(*cmd)
		header_pattern = re.compile('(?P<header>\w+?):\s+(?P<value>.*)')
		match_res = map(header_pattern.match, res.splitlines())
		matched_lines = filter(None, match_res)
		matches = (match.groupdict() for match in matched_lines)
		for match in matches:
			if match['header'] == 'changeset':
				id, sep, rev = match['value'].partition(':')
			if match['header'] == 'tag':
				tag = match['value']
				yield TaggedRevision(tag, rev)

	def _get_rev_num(self, rev=None):
		"""
		Determine the revision number for a given revision specifier.
		"""
		# first, determine the numeric ID
		cmd = ['identify', '--num']
		# workaround for #4
		cmd.extend(['--config', 'defaults.identify='])
		if rev:
			cmd.extend(['--rev', rev])
		res = self._invoke(*cmd)
		return res.strip()

	def _get_tags_by_num(self):
		"""
		Return a dictionary mapping revision number to tags for that number.
		"""
		by_revision = operator.attrgetter('revision')
		tags = sorted(self.get_tags(), key=by_revision)
		revision_tags = itertools.groupby(tags, key=by_revision)
		get_id = lambda rev: rev.split(':', 1)[0]
		return dict(
			(get_id(rev), [tr.tag for tr in tr_list])
			for rev, tr_list in revision_tags
		)

	def get_repo_tags(self):
		lines = self._invoke('tags').splitlines()
		return (
			TaggedRevision(*line.rsplit(None, 1))
			for line in lines if line
		)

	def get_ancestral_tags(self, rev='.'):
		"""
		Like get_repo_tags, but only get those tags ancestral to the current
		changeset.
		"""
		spec = 'sort(ancestors({rev}), -date)'.format(**vars())
		return self._read_tags_for_revset(spec)

	def is_modified(self):
		out = self._invoke('status', '-mard')
		return bool(out)


class Git(Command):
	exe = 'git'
	version_pattern = r'git version (\d+\.\d+[^ ]*)'

	def is_valid(self):
		return super(Git, self).is_valid() and self.version_suitable()

	def version_suitable(self):
		req_ver = pkg_resources.parse_version('1.7.10')
		act_ver = pkg_resources.parse_version(self.version())
		return act_ver >= req_ver

	def find_root(self):
		try:
			return self._invoke('rev-parse', '--top-level').strip()
		except Exception:
			pass

	def find_files(self):
		all_files = self._invoke('ls-files').splitlines()
		return all_files

	def get_tags(self, rev=None):
		"""
		Return the tags for the current revision as a set
		"""
		rev = rev or 'HEAD'
		return set(self._invoke('tag', '--points-at', rev).splitlines())

	def get_repo_tags(self):
		cmd = ["for-each-ref", "--sort=-committerdate",
				"--format=%(refname:short) %(objectname:short)", "refs/tags"]

		lines = self._invoke(*cmd).splitlines()
		return (
			TaggedRevision(*line.rsplit(None, 1))
			for line in lines if line
		)

	def is_modified(self):
		"""
		Is the current state modified? (currently stubbed assuming no)
		"""
		return False



import os

import pytest

from hgtools import managers

def _ensure_present(mgr):
	try:
		mgr.version()
	except Exception:
		pytest.skip()

@pytest.fixture
def hg_repo(tmpdir):
	tmpdir.chdir()
	mgr = managers.MercurialManager()
	_ensure_present(mgr)
	mgr._invoke('init', '.')
	os.makedirs('bar')
	touch('bar/baz')
	mgr._invoke('addremove')
	mgr._invoke('ci', '-m', 'committed')
	with open('bar/baz', 'w') as baz:
		baz.write('content')
	mgr._invoke('ci', '-m', 'added content')
	return tmpdir


@pytest.fixture
def git_repo(tmpdir):
	tmpdir.chdir()
	mgr = managers.GitManager()
	_ensure_present(mgr)
	mgr._invoke('init')
	mgr._invoke('config', 'user.email', 'hgtools@example.com')
	mgr._invoke('config', 'user.name', 'HGTools')
	os.makedirs('bar')
	touch('bar/baz')
	mgr._invoke('add', '.')
	mgr._invoke('commit', '-m', 'committed')
	with open('bar/baz', 'w') as baz:
		baz.write('content')
	mgr._invoke('commit', '-am', 'added content')
	return tmpdir


def touch(filename):
	with open(filename, 'a'):
		pass



from __future__ import absolute_import

import os
import subprocess

from . import base
from . import cmd


class Subprocess(object):
	env = None

	def _invoke(self, *params):
		"""
		Invoke self.exe as a subprocess
		"""
		cmd = [self.exe] + list(params)
		proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
			stderr=subprocess.PIPE, cwd=self.location, env=self.env)
		stdout, stderr = proc.communicate()
		if not proc.returncode == 0:
			raise RuntimeError(stderr.strip() or stdout.strip())
		return stdout.decode('utf-8')


class MercurialManager(Subprocess, cmd.Mercurial, base.RepoManager):
	"""
	A RepoManager implemented by calling into the 'hg' command-line
	as a subprocess.
	"""
	priority = 1
	if os.path.isdir('.hg'):
		priority += 1

	@property
	def env(self):
		"""
		Return an environment safe for calling an `hg` subprocess.

		Removes MACOSX_DEPLOYMENT_TARGET from the env, as if there's a
		mismatch between the local Python environment and the environment
		in which `hg` is installed, it will cause an exception. See
		https://bitbucket.org/jaraco/hgtools/issue/7 for details.
		"""
		env = os.environ.copy()
		env.pop('MACOSX_DEPLOYMENT_TARGET', None)
		return env


class GitManager(Subprocess, cmd.Git, base.RepoManager):
	"""
	A RepoManager implemented by calling into the 'git' command-line
	as a subprocess.
	"""
	priority = 1
	if os.path.isdir('.git'):
		priority += 1



from __future__ import absolute_import

import sys

from . import base
from . import cmd
from . import reentry

class MercurialInProcManager(cmd.Mercurial, base.RepoManager):
	"""
	A RepoManager implemented by invoking the hg command in-process.
	"""

	def _invoke(self, *params):
		"""
		Run the self.exe command in-process with the supplied params.
		"""
		cmd = [self.exe, '-R', self.location] + list(params)
		with reentry.in_process_context(cmd) as result:
			sys.modules['mercurial.dispatch'].run()
		stdout = result.stdio.stdout.getvalue()
		stderr = result.stdio.stderr.getvalue()
		if not result.returncode == 0:
			raise RuntimeError(stderr.strip() or stdout.strip())
		return stdout.decode('utf-8')



import os
import operator

import pytest

from hgtools import managers
from hgtools.managers import subprocess

def test_subprocess_manager_invalid_when_exe_missing():
	"""
	If the hg executable dosen't exist, the manager should report
	False for .is_valid().
	"""
	non_existent_exe = '/non_existent_executable'
	assert not os.path.exists(non_existent_exe)
	mgr = subprocess.MercurialManager()
	mgr.exe = non_existent_exe
	assert not mgr.is_valid()

@pytest.mark.usefixtures("hg_repo", scope='function')
class TestRelativePaths(object):
	"""
	Issue #9 demonstrated that the Manager would inadvertently return too many
	files when location is not the root of the repo. This test demonstrates
	the expected behavior.
	"""
	def test_nested_child(self):
		test_mgr = managers.MercurialManager('.')
		assert test_mgr.find_files() == [os.path.join('bar', 'baz')]

	def test_manager_in_child(self):
		test_mgr = managers.MercurialManager('bar')
		assert test_mgr.find_files() == ['baz']

	def test_current_dir_in_child(self):
		os.chdir('bar')
		test_mgr = managers.MercurialManager('.')
		assert test_mgr.find_files() == ['baz']

@pytest.mark.usefixtures("hg_repo", scope='function')
class TestTags(object):
	def setup_method(self, method):
		self.mgr = managers.MercurialManager('.')

	def teardown_method(self, method):
		del self.mgr

	def test_single_tag(self):
		self.mgr._invoke('tag', '1.0')
		assert self.mgr.get_tags() == set(['tip'])
		self.mgr._invoke('update', '1.0')
		assert self.mgr.get_tags() == set(['1.0'])

	def test_no_tags(self):
		"No tag should return empty set"
		assert self.mgr.get_tags('0') == set([])

	def test_local_modifications(self):
		"Local modifications should return empty set"
		with open('bar/baz', 'w') as f:
			f.write('changed')
		assert self.mgr.get_tags() == set([])

	def test_parent_tag(self):
		self.mgr._invoke('tag', '1.0')
		assert self.mgr.get_tags() == set(['tip'])
		assert self.mgr.get_parent_tags() == set(['tip'])
		assert self.mgr.get_parent_tags('.') == set(['1.0'])
		assert self.mgr.get_parent_tags('tip') == set(['1.0'])
		self.mgr._invoke('tag', '1.1')
		assert self.mgr.get_tags() == set(['tip'])
		assert self.mgr.get_parent_tags() == set(['tip'])
		assert self.mgr.get_parent_tags('.') == set(['1.1'])
		assert self.mgr.get_parent_tags('tip') == set(['1.1'])

	def test_two_tags_same_revision(self):
		"""
		Always return the latest tag for a given revision
		"""
		self.mgr._invoke('tag', '1.0')
		self.mgr._invoke('tag', '-r', '1.0', '1.1')
		self.mgr._invoke('update', '1.0')
		assert set(self.mgr.get_tags()) == set(['1.0', '1.1'])

	def test_two_tags_same_revision_lexicographically_earlier(self):
		"""
		Always return the latest tag for a given revision
		"""
		self.mgr._invoke('tag', '1.9')
		self.mgr._invoke('tag', '-r', '1.9', '1.10')
		self.mgr._invoke('update', '1.9')
		assert set(self.mgr.get_tags()) == set(['1.9', '1.10'])

	def _setup_branchy_tags(self):
		"""
		Create two heads, one which has a 1.0 tag and a different one which
		has a 1.1 tag.
		"""
		# create a commit with a tag
		with open('bar/baz', 'a') as f:
			f.write('\nmore to say\n')
		self.mgr._invoke('commit', '-m', 'Had more to say')
		self.mgr._invoke('tag', '1.0')
		# update to the pre-tagged revision
		self.mgr._invoke('update', '1')
		# Make a different commit
		with open('bar/baz', 'a') as f:
			f.write('\na different concept\n')
		self.mgr._invoke('commit', '-m', 'A different approach')
		self.mgr._invoke('tag', '1.1')

	def test_ancestral_tags_local(self):
		"""
		get_ancestral_tags should only return tagged revisions ancestral
		to the current revision.
		"""
		self._setup_branchy_tags()
		tag_revs = self.mgr.get_ancestral_tags()
		tags = map(operator.attrgetter('tag'), tag_revs)
		# Two tags expected, one is the specified tag and
		# the other is 'tip'. On fast hardware, the
		# timestamps for these may be close enough that
		# mercurial doesn't sort them properly by time,
		# so normalize order.
		assert sorted(tags) == ['1.1', 'tip']

	def test_ancestral_tags_specified(self):
		"""
		get_ancestral_tags should only return tagged revisions ancestral
		to the specified revision.
		"""
		self._setup_branchy_tags()
		tag, = self.mgr.get_ancestral_tags(3)
		assert tag.tag == '1.0'



from __future__ import print_function

import sys

import py.test

from hgtools.managers import reentry

def hello_world():
	print("hello world")

def hello_unicode_world():
	print(b"hello world".decode('ascii'))

def exit_zero():
	raise SystemExit(0)

def exit_one():
	raise SystemExit(1)

def exit_string():
	raise SystemExit('who does this?')

def echo():
	print("args are", sys.argv)

class TestReEntry(object):
	"""
	Test the in-process entry point launcher
	"""
	def test_hello_world(self):
		with reentry.in_process_context(['hello-world.py']) as proc:
			hello_world()
		assert proc.returncode == 0
		assert proc.stdio.stdout.getvalue() == 'hello world\n'

	def test_hello_world_unicode(self):
		with reentry.in_process_context(['hello-world.py']) as proc:
			hello_unicode_world()
		assert proc.returncode == 0
		assert proc.stdio.stdout.getvalue() == 'hello world\n'

	def test_main_with_system_exit(self):
		with reentry.in_process_context(['exit-zero']) as proc:
			exit_zero()
		assert proc.returncode == 0
		assert proc.stdio.stdout.getvalue() == ''

	def test_main_with_system_exit_one(self):
		with reentry.in_process_context(['exit-one']) as proc:
			exit_one()
		assert proc.returncode == 1
		assert proc.stdio.stdout.getvalue() == ''

	def test_main_with_system_exit_string(self):
		with reentry.in_process_context(['exit-string']) as proc:
			exit_string()
		assert proc.returncode == 1
		assert proc.stdio.stdout.getvalue() == ''

	def test_echo_args(self):
		with reentry.in_process_context(['echo', 'foo', 'bar']) as proc:
			echo()
		assert proc.returncode == 0
		out = "args are ['echo', 'foo', 'bar']\n"
		assert proc.stdio.stdout.getvalue() == out

class TestErrors(object):
	def test_name_error(self):
		with py.test.raises(NameError) as exc_info:
			with reentry.in_process_context([]) as proc:
				not_present
		assert proc.returncode == 1
		msg = "name 'not_present' is not defined"
		assert str(exc_info.value).endswith(msg)

	def test_keyboard_interrupt(self):
		with py.test.raises(KeyboardInterrupt):
			with reentry.in_process_context([]) as proc:
				raise KeyboardInterrupt()
		assert proc.returncode == 1



"""
setuptools plugins
"""

import sys
try:
	# Prefer the Python 2 version of configparser because Python 3
	#  backport has issues when non-unicode strings are passed.
	# See issue #10 for details
	import ConfigParser as configparser
except ImportError:
	import configparser

from . import managers

__all__ = ['file_finder', 'version_calc']

def file_finder(dirname="."):
	"""
	Find the files in ``dirname`` under Mercurial version control
	according to the setuptools spec (see
	http://peak.telecommunity.com/DevCenter/setuptools#adding-support-for-other-revision-control-systems
	).
	"""
	import distutils.log
	dirname = dirname or '.'
	try:
		valid_mgrs = managers.RepoManager.get_valid_managers(dirname)
		valid_mgrs = managers.RepoManager.existing_only(valid_mgrs)
		for mgr in valid_mgrs:
			try:
				return mgr.find_all_files()
			except Exception:
				e = sys.exc_info()[1]
				distutils.log.warn("hgtools.%s could not find files: %s",
					mgr, e)
	except Exception:
		e = sys.exc_info()[1]
		distutils.log.warn("Unexpected error finding valid managers in "
			"hgtools.file_finder_plugin: %s", e)
	return []

def patch_egg_info(force_hg_version=False):
	"""
	A hack to replace egg_info.tagged_version with a wrapped version
	that will use the mercurial version if indicated.

	`force_hg_version` is used for hgtools itself.
	"""
	from setuptools.command.egg_info import egg_info
	from pkg_resources import safe_version
	import functools
	orig_ver = egg_info.tagged_version

	@functools.wraps(orig_ver)
	def tagged_version(self):
		vcs_param = (
			getattr(self.distribution, 'use_vcs_version', False) or
			getattr(self.distribution, 'use_hg_version', False)
		)
		using_hg_version = force_hg_version or vcs_param
		if force_hg_version:
			# disable patched `tagged_version` to avoid affecting
			#  subsequent installs in the same interpreter instance.
			egg_info.tagged_version = orig_ver
		if using_hg_version:
			result = safe_version(self.distribution.get_version())
		else:
			result = orig_ver(self)
		self.tag_build = result
		return result
	egg_info.tagged_version = tagged_version

def _calculate_version(mgr, options):
	default_increment = options.get('increment')
	repo_exists = bool(mgr.find_root())
	return (mgr.get_current_version(default_increment)
		if repo_exists else default_increment)

def calculate_version(options={}):
	# The version is cached in the tag_build value in setup.cfg (so that
	#  sdist packages will have a copy of the version as determined at
	#  the build environment).
	parser = configparser.ConfigParser()
	parser.read('setup.cfg')
	has_tag_build = (parser.has_section('egg_info')
		and 'tag_build' in parser.options('egg_info'))
	if has_tag_build:
		# a cached version is available, so use it.
		version = parser.get('egg_info', 'tag_build')
	else:
		# No version is stored in tag_build, so calculate
		#  the version using a RepoManager.
		mgr = managers.RepoManager.get_first_valid_manager()
		version_handler = options.get('version_handler', _calculate_version)
		try:
			version = version_handler(mgr, options)
		except Exception:
			version = options.get('increment')
	return version

def version_calc(dist, attr, value):
	"""
	Handler for parameter to setup(use_vcs_version=value)
	attr should be 'use_vcs_version' (also allows use_hg_version for
		compatibility).
	bool(value) should be true to invoke this plugin.
	value may optionally be a dict and supply options to the plugin.
	"""
	expected_attrs = 'use_hg_version', 'use_vcs_version'
	if not value or not attr in expected_attrs: return
	options = value if isinstance(value, dict) else {}
	dist.metadata.version = calculate_version(options)
	patch_egg_info()



"""
hgtools implements several repo managers, each of which provides an interface
to Mercurial functionality.
"""

import posixpath
import itertools

from hgtools import versioning
from hgtools.util import itersubclasses

class RepoManager(versioning.VersionManagement, object):
	"""
	An abstract class defining some interfaces for working with
	repositories.
	"""
	def __init__(self, location='.'):
		self.location = location
		self.setup()

	def is_valid(self):
		"Return True if this is a valid manager for this location."
		return True

	def setup(self):
		pass

	@classmethod
	def get_valid_managers(cls, location):
		"""
		Get the valid RepoManagers for this location.
		"""
		by_priority_attr = lambda c: getattr(c, 'priority', 0)
		classes = sorted(itersubclasses(cls), key=by_priority_attr,
			reverse=True)
		all_managers = (c(location) for c in classes)
		return (mgr for mgr in all_managers if mgr.is_valid())

	@staticmethod
	def get_first_valid_manager(location='.'):
		try:
			return next(RepoManager.get_valid_managers(location))
		except StopIteration as e:
			e.args = "No source repo or suitable VCS version found",
			raise

	@staticmethod
	def existing_only(managers):
		"""
		Return only those managers that refer to an existing repo
		"""
		return (mgr for mgr in managers if mgr.find_root())

	def __repr__(self):
		return '{self.__class__.__name__}({self.location})'.format(**vars())

	def find_root(self):
		raise NotImplementedError()

	def find_files(self):
		raise NotImplementedError()

	def get_tags(self, rev=None):
		"""
		Get the tags for the specified revision (or the current revision
		if none is specified).
		"""
		raise NotImplementedError()

	def get_repo_tags(self):
		"""
		Get all tags for the repository.
		"""
		raise NotImplementedError()

	def get_parent_tags(self, rev=None):
		"""
		Return the tags for the parent revision (or None if no single
			parent can be identified).
		"""
		try:
			parent_rev = one(self.get_parent_revs(rev))
		except Exception:
			return None
		return self.get_tags(parent_rev)

	def get_parent_revs(self, rev=None):
		"""
		Get the parent revision for the specified revision (or the current
		revision if none is specified).
		"""
		raise NotImplementedError

	def is_modified(self):
		'Does the current working copy have modifications'
		raise NotImplementedError()

	def find_all_files(self):
		"""
		Find files including those in subrepositories.
		"""
		files = self.find_files()
		subrepo_files = (
			posixpath.join(subrepo.location, filename)
			for subrepo in self.subrepos()
			for filename in subrepo.find_files()
		)
		return itertools.chain(files, subrepo_files)

	def subrepos(self):
		try:
			with open(posixpath.join(self.location, '.hgsub')) as file:
				subs = list(file)
		except Exception:
			subs = []

		locs = [part.partition('=')[0].strip() for part in subs]
		return [self.__class__(posixpath.join(self.location, loc)) for loc in locs]

# from jaraco.util.itertools
def one(item):
	"""
	Return the first element from the iterable, but raise an exception
	if elements remain in the iterable after the first.

	>>> one([3])
	3
	>>> one(['val', 'other'])
	Traceback (most recent call last):
	...
	ValueError: item contained more than one value
	>>> one([])
	Traceback (most recent call last):
	...
	StopIteration
	"""
	iterable = iter(item)
	result = next(iterable)
	if tuple(itertools.islice(iterable, 1)):
		raise ValueError("item contained more than one value")
	return result



from __future__ import absolute_import

import sys
import io
import collections
import contextlib

SavedIO = collections.namedtuple('SavedIO', 'stdout stderr')

text_type = __builtins__.get('unicode', str)

class TextIO(io.StringIO):
	def write(self, data):
		if not isinstance(data, text_type):
			data = text_type(data, getattr(self, '_encoding', 'UTF-8'), 'replace')
		io.StringIO.write(self, data)

@contextlib.contextmanager
def capture_stdio():
	sys_stdout, sys.stdout = sys.stdout, TextIO()
	sys_stderr, sys.stderr = sys.stderr, TextIO()
	try:
		yield SavedIO(sys.stdout, sys.stderr)
	finally:
		sys.stdout = sys_stdout
		sys.stderr = sys_stderr

@contextlib.contextmanager
def replace_sysargv(params):
	sys_argv, sys.argv = sys.argv, params
	try:
		yield
	finally:
		sys.argv = sys_argv

class Result(object):
	pass

@contextlib.contextmanager
def capture_system_exit():
	res = Result()
	try:
		yield res
		res.code = 0
	except SystemExit as e:
		if isinstance(e.code, int):
			res.code = e.code
		else:
			res.code = 1
	except BaseException:
		res.code = 1
		raise

class ProcessResult(object):
	pass

@contextlib.contextmanager
def in_process_context(params):
	res = ProcessResult()
	try:
		with capture_stdio() as stdio:
			with replace_sysargv(params):
				with capture_system_exit() as proc_res:
					yield res
	finally:
		res.stdio = stdio
		res.returncode = proc_res.code



from __future__ import division

import operator
from functools import reduce

from distutils.version import StrictVersion

def find(pred, items):
	"""
	Find the index of the first element in items for which pred returns
	True

	>>> find(lambda x: x > 3, range(100))
	4
	>>> find(lambda x: x < -3, range(100)) is None
	True
	"""
	for i, item in enumerate(items):
		if pred(item): return i

def rfind(pred, items):
	"""
	Find the index of the last element in items for which pred returns
	True. Returns a negative number useful for indexing from the end
	of a list or tuple.

	>>> rfind(lambda x: x > 3, [5,4,3,2,1])
	-4
	"""
	return -find(pred, reversed(items)) - 1

class SummableVersion(StrictVersion):
	"""
	A special version of a StrictVersion which can be added to another
	StrictVersion.

	>>> SummableVersion('1.1') + StrictVersion('2.3')
	SummableVersion ('3.4')
	"""
	def __add__(self, other):
		result = SummableVersion(str(self))
		result.version = tuple(map(operator.add, self.version, other.version))
		return result

	def reset_less_significant(self, significant_version):
		"""
		Reset to zero all version info less significant than the
		indicated version.

		>>> ver = SummableVersion('3.1.2')
		>>> ver.reset_less_significant(SummableVersion('0.1'))
		>>> str(ver)
		'3.1'
		"""
		nonzero = lambda x: x != 0
		version_len = 3  # strict versions are always a tuple of 3
		significant_pos = rfind(nonzero, significant_version.version)
		significant_pos = version_len + significant_pos + 1
		self.version = (self.version[:significant_pos]
			+ (0,) * (version_len - significant_pos))

	def as_number(self):
		"""
		>>> round(SummableVersion('1.9.3').as_number(), 12)
		1.93
		"""
		def combine(subver, ver):
			return subver / 10 + ver
		return reduce(combine, reversed(self.version))

class VersionManagement(object):
	"""
	Version functions for RepoManager classes
	"""

	increment = '0.0.1'

	@staticmethod
	def __versions_from_tags(tags):
		for tag in tags:
			try:
				yield StrictVersion(tag)
			except ValueError:
				pass

	@staticmethod
	def __best_version(versions):
		try:
			return max(versions)
		except ValueError:
			pass

	def get_strict_versions(self):
		"""
		Return all version tags that can be represented by a
		StrictVersion.
		"""
		return self.__versions_from_tags(
			tag.tag for tag in self.get_repo_tags()
		)

	def get_tagged_version(self):
		"""
		Get the version of the local working set as a StrictVersion or
		None if no viable tag exists. If the local working set is itself
		the tagged commit and the tip and there are no local
		modifications, use the tag on the parent changeset.
		"""
		tags = list(self.get_tags())
		if 'tip' in tags and not self.is_modified():
			tags = self.get_parent_tags('tip')
		versions = self.__versions_from_tags(tags)
		return self.__best_version(versions)

	def get_latest_version(self):
		"""
		Determine the latest version ever released of the project in
		the repo (based on tags).
		"""
		return self.__best_version(self.get_strict_versions())

	def get_current_version(self, increment=None):
		"""
		Return as a string the version of the current state of the
		repository -- a tagged version, if present, or the next version
		based on prior tagged releases.
		"""
		ver = (
			self.get_tagged_version()
			or str(self.get_next_version(increment)) + '.dev0'
			)
		return str(ver)

	def get_next_version(self, increment=None):
		"""
		Return the next version based on prior tagged releases.
		"""
		increment = increment or self.increment
		return self.infer_next_version(self.get_latest_version(), increment)

	@staticmethod
	def infer_next_version(last_version, increment):
		"""
		Given a simple application version (as a StrictVersion),
		and an increment (1.0, 0.1, or 0.0.1), guess the next version.

		Set up a shorthand for examples

		>>> VM_infer = lambda *params: str(VersionManagement.infer_next_version(*params))

		>>> VM_infer('3.2', '0.0.1')
		'3.2.1'
		>>> VM_infer(StrictVersion('3.2'), '0.0.1')
		'3.2.1'
		>>> VM_infer('3.2.3', '0.1')
		'3.3'
		>>> VM_infer('3.1.2', '1.0')
		'4.0'

		Subversions never increment parent versions

		>>> VM_infer('3.0.9', '0.0.1')
		'3.0.10'

		If it's a prerelease version, just remove the prerelease.

		>>> VM_infer('3.1a1', '0.0.1')
		'3.1'

		If there is no last version, use the increment itself

		>>> VM_infer(None, '0.1')
		'0.1'
		"""
		if last_version is None:
			return increment
		last_version = SummableVersion(str(last_version))
		if last_version.prerelease:
			last_version.prerelease = None
			return str(last_version)
		increment = SummableVersion(increment)
		sum = last_version + increment
		sum.reset_less_significant(increment)
		return sum
